/*
 * Copyright (c) 2020 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <zephyr/kernel.h>

#include <app_event_manager.h>

#define MODULE main
#include "module_state_event.h"

#include <zephyr/logging/log.h>
#include <zephyr/drivers/uart.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/sys/ring_buffer.h>

LOG_MODULE_REGISTER(MODULE,4);

#define USB_SERIALNUMBER_TEMPLATE "THINGY91_%04X%08X"
static uint8_t usb_serial_str[] = "THINGY91_12PLACEHLDRS";

/* Overriding weak function to set iSerialNumber at runtime. */
uint8_t *usb_update_sn_string_descriptor(void)
{
	snprintk(usb_serial_str, sizeof(usb_serial_str), USB_SERIALNUMBER_TEMPLATE,
				(uint32_t)(NRF_FICR->DEVICEADDR[1] & 0x0000FFFF)|0x0000C000,
				(uint32_t)NRF_FICR->DEVICEADDR[0]);

	return usb_serial_str;
}


#define UART_BUF_SIZE		23
#define UART_TX_TIMEOUT_MS	100
#define UART_RX_TIMEOUT_MS	100

K_SEM_DEFINE(tx_done, 1, 1);
K_SEM_DEFINE(rx_disabled, 0, 1);

#define UART_TX_BUF_SIZE  		256
#define UART_RX_MSG_QUEUE_SIZE	8
struct uart_msg_queue_item {
	uint8_t bytes[UART_BUF_SIZE];
	uint32_t length;
};

// UART TX fifo
RING_BUF_DECLARE(app_tx_fifo, UART_TX_BUF_SIZE);
volatile int bytes_claimed;

// UART RX primary buffers
uint8_t uart_double_buffer[2][UART_BUF_SIZE];
uint8_t *uart_buf_next = uart_double_buffer[1];

// UART RX message queue
K_MSGQ_DEFINE(uart_rx_msgq, sizeof(struct uart_msg_queue_item), UART_RX_MSG_QUEUE_SIZE, 4);

static const struct device *dev_uart;

uint8_t tx_buffer_1[] = "Hi From nrf52840 SoC\x1A";
uint8_t tx_buffer_2[] = "did you received the packet\x1A";
static int uart_tx_get_from_queue(void)
{
	uint8_t *data_ptr;
	// Try to claim any available bytes in the FIFO
	bytes_claimed = ring_buf_get_claim(&app_tx_fifo, &data_ptr, UART_TX_BUF_SIZE);

	if(bytes_claimed > 0) {
		// Start a UART transmission based on the number of available bytes
		uart_tx(dev_uart, data_ptr, bytes_claimed, SYS_FOREVER_MS);
	}
	return bytes_claimed;
}

void app_uart_async_callback(const struct device *uart_dev,
							 struct uart_event *evt, void *user_data)
{
	static struct uart_msg_queue_item new_message;

	switch (evt->type) {
		case UART_TX_DONE:
			// Free up the written bytes in the TX FIFO
			ring_buf_get_finish(&app_tx_fifo, bytes_claimed);
			LOG_DBG("TX done\n");
			// If there is more data in the TX fifo, start the transmission
			if(uart_tx_get_from_queue() == 0) {
				// Or release the semaphore if the TX fifo is empty
				k_sem_give(&tx_done);
			}
			break;
		
		case UART_RX_RDY:
			memcpy(new_message.bytes, evt->data.rx.buf + evt->data.rx.offset, evt->data.rx.len);
			new_message.length = evt->data.rx.len;
            LOG_DBG("new_message.length : %d\n",new_message.length);
			if(k_msgq_put(&uart_rx_msgq, &new_message, K_NO_WAIT) != 0){
				printk("Error: Uart RX message queue full!\n");
			}
			break;
		
		case UART_RX_BUF_REQUEST:
			uart_rx_buf_rsp(dev_uart, uart_buf_next, UART_BUF_SIZE);
			break;

		case UART_RX_BUF_RELEASED:
			uart_buf_next = evt->data.rx_buf.buf;
			break;

		case UART_RX_DISABLED:
			k_sem_give(&rx_disabled);
			break;
		
		default:
			break;
	}
}

static void app_uart_init(void)
{
	dev_uart = DEVICE_DT_GET(DT_NODELABEL(uart1));

    if (!dev_uart) {
        LOG_ERR("Impossible de trouver le périphérique UART\n");
        return;
    }
    if(device_is_ready(dev_uart) == false)
    {
        LOG_ERR("Périphérique UART non prêt\n");
        return -1;
    }

	uart_callback_set(dev_uart, app_uart_async_callback, NULL);
	uart_rx_enable(dev_uart, uart_double_buffer[0], UART_BUF_SIZE, UART_RX_TIMEOUT_MS);
}

void send_data_to_nrf91(const uint8_t * data_ptr, uint32_t data_len)
{
    uint32_t written_to_buf = ring_buf_put(&app_tx_fifo, data_ptr, data_len);
		
    // In case the UART TX is idle, start transmission
    if(k_sem_take(&tx_done, K_NO_WAIT) == 0) {
        LOG_DBG("Transmission started\n");
        uart_tx_get_from_queue();
    }	
    
    // In case some data is still to be written, sleep for some time and run the loop one more time
    k_msleep(3000);
    
}
// Function to send UART data, by writing it to a ring buffer (FIFO) in the application
// WARNING: This function is not thread safe! If you want to call this function from multiple threads a semaphore should be used
static int app_uart_send()
{
	while(1) {
		
        send_data_to_nrf91(tx_buffer_1, strlen(tx_buffer_1));
        k_msleep(1000);
        send_data_to_nrf91(tx_buffer_2, strlen(tx_buffer_2));
        k_msleep(3000);
		
	}
    LOG_DBG("Transmission completed\n");
	return 0;
}

int main(void)
{
	/*
	printk("Bluetooth initialized\n");

	start_scan();*/
	if (app_event_manager_init()) {
		LOG_ERR("Application Event Manager not initialized");
	} else {
		module_set_state(MODULE_STATE_READY);
	}
	
	// printk("UART Async example started\n");
	
	app_uart_init();

	const uint8_t test_string[] = "Hello world through the UART async driver\x1A";
	// app_uart_send();

	struct uart_msg_queue_item incoming_message;
    static uint8_t string_buffer[UART_BUF_SIZE + 1] = {0};
    int buffer_length = 0;

    while (1) {
        // This function will not return until a new message is ready
		LOG_DBG("Waiting for a new message\n");
        k_msgq_get(&uart_rx_msgq, &incoming_message, K_FOREVER);
		LOG_DBG("New message received\n");
        // Copy incoming_message.bytes to string_buffer
        for (int i = 0; i < incoming_message.length; i++) {
            string_buffer[buffer_length] = incoming_message.bytes[i];
            buffer_length++;
            if (incoming_message.bytes[i] == '\x1A') {
                // End of message detected
                LOG_DBG("End of message detected , string_buffer[%d] : %c \n", buffer_length - 1, string_buffer[buffer_length - 1]);
                buffer_length--; // Decrease buffer length to remove the last character
                string_buffer[buffer_length] = '\0'; // Null-terminate the string
                printk("RX : %s\n", string_buffer);
                memset(string_buffer, 0, sizeof(string_buffer)); // Clear the buffer
                buffer_length = 0; // Reset buffer length
                break;
            }
        }
    }

	
	return 0;
}